@@ -1,7 +1,8 @@
 ;;; Guile-zlib --- GNU Guile bindings of zlib
-;;; Copyright © 2016, 2017, 2021 Ludovic Courtès <ludo@gnu.org>
+;;; Copyright © 2016-2017, 2021, 2024 Ludovic Courtès <ludo@gnu.org>
 ;;; Copyright © 2020 Mathieu Othacehe <othacehe@gnu.org>
 ;;; Copyright © 2013 David Thompson <dthompson2@worcester.edu>
+;;; Copyright © 2023 Artyom V. Poptsov <poptsov.artyom@gmail.com>
 ;;;
 ;;; This file is part of Guile-zlib.
 ;;;
@@ -41,7 +42,17 @@
             make-zlib-input-port
             make-zlib-output-port
             call-with-zlib-input-port
-            call-with-zlib-output-port))
+            call-with-zlib-output-port
+
+            ;; Error codes returned together with 'zlib-error' exceptions.
+            Z_OK
+            Z_STREAM_END
+            Z_NEED_DICT
+            Z_ERRNO
+            Z_STREAM_ERROR
+            Z_DATA_ERROR
+            Z_MEM_ERROR
+            Z_BUF_ERROR))
 
 ;;; Commentary:
 ;;;
@@ -72,9 +83,21 @@ either zlib or the function could not be found."
     (format port "#<gzip-file ~a>"
             (number->string (object-address obj) 16))))
 
+;; On Guile 2.0, (set-port-encoding! PORT #f) throws.  Work around it.
+(cond-expand
+ ((not guile-2.2)                                 ;Guile 2.0
+  (define (set-port-encoding!* port encoding)
+    (when encoding
+      (set-port-encoding! port encoding))))
+ (else
+  (define set-port-encoding!* set-port-encoding!)))
+
 (define gzerror
   (let ((proc (zlib-procedure '* "gzerror" '(* *))))
     (lambda (gzfile)
+      "Return two values: the error code corresponding to the current error
+on GZFILE (one of the Z_* constants), and the corresponding error message as
+a string."
       (let* ((errnum* (make-bytevector (sizeof int)))
              (ptr     (proc (gzip-file->pointer gzfile)
                             (bytevector->pointer errnum*))))
@@ -94,6 +117,24 @@ closes FD."
             (throw 'zlib-error 'gzdopen)
             (pointer->gzip-file result))))))
 
+(define (raise-gzerror-exception proc gzfile)
+  "Raise a 'zlib-error' exception that includes information returned by
+'gzerror' on GZFILE."
+  (call-with-values (lambda () (gzerror gzfile))
+    (lambda (code message)
+      (throw 'zlib-error proc code message))))
+
+(define (print-zlib-error port key args default-printer)
+  (match args
+    ((_ proc)
+     (format port "In procedure ~a: unspecified zlib error" proc))
+    ((_ proc code)
+     (format port "In procedure ~a: zlib error ~a" proc code))
+    ((_ proc code message . _)
+     (format port "In procedure ~a: zlib error ~a: ~a" proc code message))))
+
+(set-exception-printer! 'zlib-error print-zlib-error)
+
 (define gzread!
   (let ((proc (zlib-procedure int "gzread" (list '* '* unsigned-int))))
     (lambda* (gzfile bv #:optional (start 0) (count (bytevector-length bv)))
@@ -104,7 +145,7 @@ the end-of-stream has been reached."
                        (bytevector->pointer bv start)
                        count)))
         (if (< ret 0)
-            (throw 'zlib-error 'gzread! ret)
+            (raise-gzerror-exception 'gzread! gzfile)
             ret)))))
 
 (define gzwrite
@@ -116,7 +157,7 @@ the number of uncompressed bytes written, a strictly positive integer."
                        (bytevector->pointer bv start)
                        count)))
         (if (<= ret 0)
-            (throw 'zlib-error 'gzwrite ret)
+            (raise-gzerror-exception 'gzwrite gzfile)
             ret)))))
 
 (define gzbuffer!
@@ -125,7 +166,7 @@ the number of uncompressed bytes written, a strictly positive integer."
       "Change the internal buffer size of GZFILE to SIZE bytes."
       (let ((ret (proc (gzip-file->pointer gzfile) size)))
         (unless (zero? ret)
-          (throw 'zlib-error 'gzbuffer! ret))))))
+          (raise-gzerror-exception 'gzbuffer gzfile))))))
 
 (define gzeof?
   (let ((proc (zlib-procedure int "gzeof" '(*))))
@@ -139,7 +180,7 @@ the number of uncompressed bytes written, a strictly positive integer."
       "Close GZFILE."
       (let ((ret (proc (gzip-file->pointer gzfile))))
         (unless (zero? ret)
-          (throw 'zlib-error 'gzclose ret (gzerror gzfile)))))))
+          (raise-gzerror-exception 'gzclose gzfile))))))
 
 
 
@@ -156,11 +197,12 @@ the number of uncompressed bytes written, a strictly positive integer."
   -1)
 
 (define* (make-gzip-input-port port #:key (buffer-size %default-buffer-size))
-  "Return an input port that decompresses data read from PORT, a file port.
-PORT is automatically closed when the resulting port is closed.  BUFFER-SIZE
-is the size in bytes of the internal buffer, 8 KiB by default; using a larger
-buffer increases decompression speed.  An error is thrown if PORT contains
-buffered input, which would be lost (and is lost anyway)."
+  "Return an input port that decompresses data read from PORT, a file port.  The
+new port inherits the encoding of PORT.  PORT is automatically closed when the
+resulting port is closed.  BUFFER-SIZE is the size in bytes of the internal
+buffer, 8 KiB by default; using a larger buffer increases decompression speed.
+An error is thrown if PORT contains buffered input, which would be lost (and is
+lost anyway)."
   (define gzfile
     (match (drain-input port)
       (""                                         ;PORT's buffer is empty
@@ -182,21 +224,26 @@ buffered input, which would be lost (and is lost anyway)."
   (define (read! bv start count)
     (gzread! gzfile bv start count))
 
+  (define encoding (port-encoding port))
+
   (unless (= buffer-size %default-buffer-size)
     (gzbuffer! gzfile buffer-size))
 
   (close-port port)                               ;we no longer need it
-  (make-custom-binary-input-port "gzip-input" read! #f #f
-                                 (lambda ()
-                                   (gzclose gzfile))))
+
+  (let ((result (make-custom-binary-input-port "gzip-input" read! #f #f
+                                               (lambda () (gzclose gzfile)))))
+    (set-port-encoding!* result encoding)
+    result))
+
 
 (define* (make-gzip-output-port port
                                 #:key
                                 (level %default-compression-level)
                                 (buffer-size %default-buffer-size))
-  "Return an output port that compresses data at the given LEVEL, using PORT,
-a file port, as its sink.  PORT must be a file port; it is automatically
-closed when the resulting port is closed."
+  "Return an output port that compresses data at the given LEVEL, using PORT, a
+file port, as its sink.  PORT must be a file port; it is automatically closed
+when the resulting port is closed.  The new port inherits the encoding of PORT."
   (define gzfile
     (begin
       (force-output port)                         ;empty PORT's buffer
@@ -206,49 +253,57 @@ closed when the resulting port is closed."
   (define (write! bv start count)
     (gzwrite gzfile bv start count))
 
+  (define encoding (port-encoding port))
+
   (unless (= buffer-size %default-buffer-size)
     (gzbuffer! gzfile buffer-size))
 
   (close-port port)
-  (make-custom-binary-output-port "gzip-output" write! #f #f
-                                  (lambda ()
-                                    (gzclose gzfile))))
+  (let ((result (make-custom-binary-output-port "gzip-output" write! #f #f
+                                                (lambda () (gzclose gzfile)))))
+    (set-port-encoding!* result encoding)
+    result))
+
+(define (call-with-zlib-port port proc)
+  "Call PROC with PORT as its argument.  Close PORT when leaving the dynamic
+extent of PROC via return or exception."
+  ;; Note: Not using 'dynamic-wind' to allow for 'abort-to-prompt' calls in
+  ;; PROC, as happens with Fibers.
+  (catch #t
+    (lambda ()
+      (let ((result (proc port)))
+        (close-port port)
+        result))
+    (lambda args
+      (close-port port)
+      (apply throw args))))
 
 (define* (call-with-gzip-input-port port proc
                                     #:key (buffer-size %default-buffer-size))
   "Call PROC with a port that wraps PORT and decompresses data read from it.
-PORT must be a file port; it is closed upon completion.  The gzip internal
-buffer size is set to BUFFER-SIZE bytes.
+PORT must be a file port; it is closed upon completion.  The new port inherits
+the encoding of PORT.  The gzip internal buffer size is set to BUFFER-SIZE
+bytes.
 
 See 'call-with-zlib-input-port' for a slightly slower variant that does not
 require PORT to be a file port."
-  (let ((gzip (make-gzip-input-port port #:buffer-size buffer-size)))
-    (dynamic-wind
-      (const #t)
-      (lambda ()
-        (proc gzip))
-      (lambda ()
-        (close-port gzip)))))
+  (call-with-zlib-port (make-gzip-input-port port #:buffer-size buffer-size)
+    proc))
 
 (define* (call-with-gzip-output-port port proc
                                      #:key
                                      (level %default-compression-level)
                                      (buffer-size %default-buffer-size))
-  "Call PROC with an output port that wraps PORT and compresses data.  PORT
-must be a file port; it is closed upon completion.  The gzip internal buffer
-size is set to BUFFER-SIZE bytes.
+  "Call PROC with an output port that wraps PORT and compresses data.  PORT must
+be a file port; it is closed upon completion.  The new port inherits the
+encoding of PORT.  The gzip internal buffer size is set to BUFFER-SIZE bytes.
 
 See 'call-with-zlib-output-port' for a slightly slower variant that does not
 require PORT to be a file port."
-  (let ((gzip (make-gzip-output-port port
-                                     #:level level
-                                     #:buffer-size buffer-size)))
-    (dynamic-wind
-      (const #t)
-      (lambda ()
-        (proc gzip))
-      (lambda ()
-        (close-port gzip)))))
+  (call-with-zlib-port (make-gzip-output-port port
+                                              #:level level
+                                              #:buffer-size buffer-size)
+    proc))
 
 
 ;;;
@@ -279,6 +334,15 @@ require PORT to be a file port."
 (define %uncompress
   (zlib-procedure int "uncompress" (list '* '* '* unsigned-long)))
 
+;; ZEXTERN int ZEXPORT uncompress2(Bytef *dest,   uLongf *destLen,
+;;                                 const Bytef *source, uLong *sourceLen);
+;;
+;; Same as uncompress, except that sourceLen is a pointer, where the length of
+;; the source is *sourceLen. On return, *sourceLen is the number of source bytes
+;; consumed.
+(define %uncompress2
+  (zlib-procedure int "uncompress2" (list '* '* '* '*)))
+
 ;;
 ;; ZEXTERN int ZEXPORT compress OF((Bytef *dest, uLongf *destLen,
 ;;                                  const Bytef *source, uLong sourceLen));
@@ -361,34 +425,39 @@ require PORT to be a file port."
 the uncompressed data."
   (define (try-uncompress length)
     (let* ((dest (make-bytevector (* (sizeof uint8) length)))
-           (dest-length (make-bytevector (sizeof unsigned-long))))
+           (dest-length   (make-bytevector 8))
+           (source-length (make-bytevector 8)))
+      (bytevector-u64-set! source-length 0 length (native-endianness))
       (bytevector-uint-set! dest-length 0 length
                             (native-endianness) (sizeof unsigned-long))
-      (values (%uncompress (bytevector->pointer dest)
-                   (bytevector->pointer dest-length)
-                   (bytevector->pointer bv)
-                   length)
-              (bytevector-copy-region dest 0 (buffer-length dest-length)))))
+      (let ((result (%uncompress2 (bytevector->pointer dest)
+                                  (bytevector->pointer dest-length)
+                                  (bytevector->pointer bv)
+                                  (bytevector->pointer source-length))))
+        (values result
+                (bytevector-copy-region dest 0 (buffer-length dest-length))
+                (buffer-length source-length)))))
 
-  ;; We don't know how much space we need to store the uncompressed
-  ;; data. So, we make an initial guess and keep increasing buffer
-  ;; size until it works.
-  (define (step-buffer-length length)
-    (inexact->exact (round (* length 1.5))))
-
-  (let try-again ((tries 1)
-                  (length (step-buffer-length (bytevector-length bv))))
+  (let try-again ((tries  1)
+                  (length (inexact->exact
+                           (round (* (bytevector-length bv) 1.5)))))
     ;; Bail after so many failed attempts. This shouldn't happen, but
     ;; I don't like the idea of a potentially unbounded loop that
     ;; keeps allocating larger and larger chunks of memory.
     (if (> tries 10)
         (throw 'zlib-error 'uncompress 0)
-        (receive (ret-code uncompressed-data)
+        (receive (ret-code uncompressed-data read-source-length)
             (try-uncompress length)
           ;; return code -5 means that destination buffer was too small.
           ;; return code  0 means everything went OK.
           (cond ((= ret-code -5)
-                 (try-again (1+ tries) (step-buffer-length length)))
+                 (let* ((compressesed-data-size (bytevector-length bv))
+                        (percent                (* (/ read-source-length
+                                                      compressesed-data-size)
+                                                   100))
+                        (next-size (inexact->exact (round (/ (* 100 length)
+                                                             percent)))))
+                   (try-again (1+ tries) next-size)))
                 ((= ret-code 0)
                  uncompressed-data)
                 (else
@@ -584,8 +653,8 @@ struct's fields."
                                (buffer-size %default-buffer-size)
                                (close? #t))
   "Return an input port that decompresses data read from PORT.  FORMAT is a
-symbol denoting the header format; it must be one of 'deflate (RFC 1950),
-'zlib (RFC 1951), or 'gzip (RFC 1952).
+symbol denoting the header format; it must be one of 'deflate (RFC 1950), 'zlib
+(RFC 1951), or 'gzip (RFC 1952).  The new port inherits the encoding of PORT.
 
 When CLOSE? is true, PORT is automatically closed when the resulting port is
 closed."
@@ -654,6 +723,8 @@ closed."
                           (throw 'zlib-error ret
                                  (stream-error-message* stream))))))))))
 
+  (define encoding (port-encoding port))
+
   (define result
     (make-custom-binary-input-port "zlib-input" read! #f #f
                                    (lambda ()
@@ -669,6 +740,7 @@ closed."
   (inflate-init! (pointer stream)
                  (window-bits-for-format format))
   (set-stream-avail-in! stream 0)
+  (set-port-encoding!* result encoding)
   result)
 
 (define* (make-zlib-output-port port
@@ -677,9 +749,10 @@ closed."
                                 (buffer-size %default-buffer-size)
                                 (level %default-compression-level)
                                 (close? #t))
-  "Return an output port that compresses data at the given LEVEL, using PORT
-as its sink.  FORMAT is a symbol denoting the header format; it must be one
-of 'deflate (RFC 1950), 'zlib (RFC 1951), or 'gzip (RFC 1952).
+  "Return an output port that compresses data at the given LEVEL, using PORT as
+its sink.  FORMAT is a symbol denoting the header format; it must be one of
+'deflate (RFC 1950), 'zlib (RFC 1951), or 'gzip (RFC 1952).  The new port
+inherits the encoding of PORT.
 
 When FORMAT is 'gzip, the gzip header takes default values, and in particular
 no modification time and no file name.
@@ -738,6 +811,8 @@ closed."
     (when close?
       (close-port port)))
 
+  (define encoding (port-encoding port))
+
   (deflate-init! (pointer stream) level
     #:window-bits (window-bits-for-format format))
 
@@ -745,7 +820,9 @@ closed."
   (set-stream-next-out! stream
                         (pointer-address (bytevector->pointer output-buffer)))
 
-  (make-custom-binary-output-port "zlib-output" write! #f #f close))
+  (let ((result (make-custom-binary-output-port "zlib-output" write! #f #f close)))
+    (set-port-encoding!* result encoding)
+    result))
 
 (define* (call-with-zlib-input-port port proc
                                     #:key
@@ -754,16 +831,11 @@ closed."
   "Call PROC with a port that wraps PORT and decompresses data read from it.
 PORT is closed upon completion.  The zlib internal buffer size is set to
 BUFFER-SIZE bytes."
-  (let ((zlib (make-zlib-input-port port
-                                    #:format format
-                                    #:buffer-size buffer-size
-                                    #:close? #t)))
-    (dynamic-wind
-      (const #t)
-      (lambda ()
-        (proc zlib))
-      (lambda ()
-        (close-port zlib)))))
+  (call-with-zlib-port (make-zlib-input-port port
+                                             #:format format
+                                             #:buffer-size buffer-size
+                                             #:close? #t)
+    proc))
 
 (define* (call-with-zlib-output-port port proc
                                      #:key
@@ -773,16 +845,13 @@ BUFFER-SIZE bytes."
   "Call PROC with an output port that wraps PORT and compresses data in the
 given FORMAT, with the given LEVEL.  PORT is closed upon completion.  The
 zlib internal buffer size is set to BUFFER-SIZE bytes."
-  (let ((zlib (make-zlib-output-port port
-                                     #:format format
-                                     #:level level
-                                     #:buffer-size buffer-size
-                                     #:close? #t)))
-    (dynamic-wind
-      (const #t)
-      (lambda ()
-        (proc zlib))
-      (lambda ()
-        (close-port zlib)))))
+  (call-with-zlib-port (make-zlib-output-port port
+                                              #:format format
+                                              #:level level
+                                              #:buffer-size buffer-size
+                                              #:close? #t)
+    proc))
 
-;;; zlib.scm ends here
+;; Local Variables:
+;; eval: (put 'call-with-zlib-port 'scheme-indent-function 1)
+;; End:
